import lodash from "lodash";
const { merge, cloneDeep, has, set } = lodash;
import { createBreakpoints } from "./breakpoints.mjs";
import { expandToEndpoints } from "./expand.mjs";
import { convertToArrays, mergeKeys, mo, ot, pipe } from "./index.mjs";
import { inspect } from "util";
import { isEndpoint } from "../resolve/endpoints.mjs";

const insertScheme = (value, ref, scheme) => {
  const r = new RegExp(`\\{${ref}\\.`, "g");
  return value.replace(r, `{${scheme}.${ref}.`);
};

const inheritScheme = (node, scheme) => {
  if (!node) {
    console.log(`ERROR: no node for inheritScheme`);
    process.exit();
  }
  const name = Object.keys(scheme)[0];
  const tokens = Object.values(scheme)[0];
  return mo(node, (node) => {
    if (ot(node) !== "string") return node;
    const match = node.match(/\{([^\.]+)\./);
    if (!match) return node;
    const ref = match[1];
    if (!tokens.hasOwnProperty(ref)) return node;
    return insertScheme(node, ref, name);
  });
};

export const inheritEndpoints = (node, fromNode) => {
  const toNode = cloneDeep(node);
  mo(fromNode, (from, path) => {
    if (isEndpoint(from) && !has(toNode, path)) {
      // console.log(`tablet not has ${path}`);
      set(toNode, path, cloneDeep(from));
    }
    return from;
  });
  return toNode;
};

const buildTree = (tokens) => {
  // console.log(`==== build tree ====`);

  const schemes = {
    light: tokens.light,
    dark: merge({}, tokens.light, tokens.dark),
  };

  // console.log(`schemes`, schemes);

  const devices = {
    phone: tokens.device.phone,
    tablet: merge({}, tokens.device.phone, tokens.device.tablet),
  };

  const components = {
    phone: tokens.component.phone,
    // tablet: tokens.component.tablet,
    tablet: inheritEndpoints(tokens.component.tablet, tokens.component.phone),
    // tablet: merge({}, tokens.component.phone, tokens.component.tablet),
  };

  // console.log(inspect(components.tablet.Link, { depth: 100 }));

  const phone = {
    ...devices.phone,
    light: inheritScheme(components.phone, { light: schemes.light }),
  };
  phone.dark = merge({}, phone.light,
    inheritScheme(components.phone, { dark: schemes.dark })
  );

  const tablet = {
    ...devices.tablet,
    light: inheritScheme(components.tablet, { light: schemes.light }),
  };

  tablet.dark = merge({}, tablet.light,
    inheritScheme(components.tablet, { dark: schemes.dark }),
  );

  const document = tokens.component.document && {
    light: inheritScheme(tokens.component.document, { light: schemes.light }),
    dark: inheritScheme(tokens.component.document, { dark: schemes.dark }),
  };

  if (document) {
    phone.light.document = document.light;
    phone.dark.document = document.dark;
    tablet.light.document = document.light;
    tablet.dark.document = document.dark;
  }

  const tree = {
    ...tokens.global,
    ...schemes,
    phone,
    tablet,
  };

  // console.log(`TREE ----------------------\n`, inspect(tree.tablet.light.component.Link.today.long, { depth: 100 }));

  return tree;
};

let id = 0;
export const makeIds = (node) =>
  mo(node, (node) => {
    if (isEndpoint(node)) {
      id++;
      return {
        ...node,
        _id: `node${id}`,
      };
    }
    return node;
  });

export const removeUnderscored = (node) =>
  mo(node, (node) => {
    const keys =
      ot(node) === "object" && Object.keys(node).some((key) => /^_/.test(key));
    if (keys)
      return Object.fromEntries(
        Object.entries(node).filter(([key]) => !/^_/.test(key))
      );
    return node;
  });

export const om = (node, breakpoints) => {
  // console.log("OM");

  const tokens = pipe(
    buildTree,
    expandToEndpoints,
    (node) => createBreakpoints(node, breakpoints),
    convertToArrays,
    removeUnderscored,
    makeIds
  )(node);

  // console.log(`TOKENS ----------------------\n`, inspect(tokens.tablet.light.component.Link.today.long, { depth: 100 }));

  return tokens;

  /* return [
    ["phone", "tablet"],
    ["light", "dark"],
    ["phone.light", "phone.dark"],
    ["tablet.light", "tablet.dark"],
  ].reduce((tokens, keys) => mergeKeys(tokens, keys), tokens); */
};

export const sortStyles = (node) =>
  mo(node, (node) => {
    if (ot(node) === "object") {
      const sorted = Object.fromEntries(
        Object.entries(node).sort((a, b) => a[0].localeCompare(b[0]))
      );
      return sorted;
    }
    return node;
  });

